"""Reading log check-ins handler and services.
"""
import json
import web

from typing import Optional

from infogami.utils import delegate
from infogami.utils.view import render_template

from openlibrary.accounts import get_current_user
from openlibrary.utils import extract_numeric_id_from_olid
from openlibrary.core.bookshelves_events import BookshelvesEvents, BookshelfEvent
from openlibrary.utils.decorators import authorized_for


def make_date_string(year: int, month: Optional[int], day: Optional[int]) -> str:
    """Creates a date string in 'YYYY-MM-DD' format, given the year, month, and day.

    Month and day can be None.  If the month is None, only the year is returned.
    If there is a month but day is None, the year and month are returned.
    """
    result = f'{year}'
    if month:
        result += f'-{month:02}'
        if day:
            result += f'-{day:02}'
    return result


class check_ins(delegate.page):
    path = r'/check-ins/OL(\d+)W'

    @authorized_for('/usergroup/admin')
    def GET(self, work_id):
        return render_template('check_ins/test_form')

    @authorized_for('/usergroup/admin')
    def POST(self, work_id):
        """Creates a check-in for the given work.

        Additional data is expected to be sent as JSON in the body, and will
        have the following keys:
        edition_olid : str,
        event_type : str,
        year : integer,
        month : integer [optional],
        day : integer [optional]
        """
        data = json.loads(web.data())
        valid_request = self.is_valid(data)
        user = get_current_user()
        username = user['key'].split('/')[-1]

        if valid_request and username:
            edition_id = extract_numeric_id_from_olid(data['edition_olid'])
            date_str = make_date_string(
                data['year'], data.get('month', None), data.get('day', None)
            )
            event_type = BookshelfEvent[data['event_type']].value
            BookshelvesEvents.create_event(
                username, work_id, edition_id, date_str, event_type=event_type
            )
        else:
            return web.badrequest(message="Invalid request")
        return delegate.RawText(json.dumps({'status': 'ok'}))

    def is_valid(self, data: dict) -> bool:
        """Validates POSTed check-in data."""
        if not all(key in data for key in ('edition_olid', 'year', 'event_type')):
            return False
        if not BookshelfEvent.has_key(data['event_type']):  # noqa: W601
            return False
        return True


class patron_check_ins(delegate.page):
    path = r'/people/([^/]+)/checkins'
    encoding = 'json'

    @authorized_for('/usergroup/admin')
    def POST(self, username):
        data = json.loads(web.data())

        if not self.is_valid(data):
            return web.badrequest(message="Invalid request")

        results = BookshelvesEvents.select_by_id(data['id'])
        if not results:
            return web.badrequest(message="Invalid request")

        row = results[0]
        if row['username'] != username:  # Cannot update someone else's records
            return web.badrequest(message="Invalid request")

        updates = {}
        if 'year' in data:
            event_date = make_date_string(
                data['year'], data.get('month', None), data.get('day', None)
            )
            updates['event_date'] = event_date

        if 'data' in data:
            updates['data'] = json.dumps(data['data'])

        records_updated = BookshelvesEvents.update_event(data['id'], **updates)

        return delegate.RawText(
            json.dumps({'status': 'success', 'updatedRecords': records_updated})
        )

    def is_valid(self, data):
        """Validates data POSTed to this handler.

        A request is invalid if it is:
        a. Missing an 'id'
        b. Does not have either 'year' or 'data'
        """
        return 'id' in data and 'data' in data and 'year' in data


def setup():
    pass
